
package cn.zijinshi.cvs.core;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.team.internal.ccvs.core.CVSException;
import org.eclipse.team.internal.ccvs.core.CVSMessages;
import org.eclipse.team.internal.ccvs.core.CVSStatus;
import org.eclipse.team.internal.ccvs.core.CVSTag;
import org.eclipse.team.internal.ccvs.core.ICVSFolder;
import org.eclipse.team.internal.ccvs.core.ICVSResource;
import org.eclipse.team.internal.ccvs.core.ICVSRunnable;
import org.eclipse.team.internal.ccvs.core.Policy;
import org.eclipse.team.internal.ccvs.core.client.Command;
import org.eclipse.team.internal.ccvs.core.client.ConsoleListeners;
import org.eclipse.team.internal.ccvs.core.client.Session;
import org.eclipse.team.internal.ccvs.core.client.Tag;
import org.eclipse.team.internal.ccvs.core.client.Command.GlobalOption;
import org.eclipse.team.internal.ccvs.core.client.listeners.ICommandOutputListener;

import cn.zijinshi.cvs.core.client.CusTagFileSender;





/**
 * 
 */

public class CTag extends Tag
{
	public static final CLocalOption C_REMOVE_TAG = new CLocalOption("-d"); //$NON-NLS-1$
	
	public static final CLocalOption C_RECURSE = new CLocalOption("-R"); //$NON-NLS-1$
	public static final CLocalOption C_DO_NOT_RECURSE = new CLocalOption("-l"); //$NON-NLS-1$	
	
	public static final CLocalOption C_CREATE_BRANCH = new CLocalOption("-b", null);	 //$NON-NLS-1$	
	/**
	 * Option subtype for local options that vary from command to command.
	 */
	public static class CLocalOption extends Option {
		protected CLocalOption(String option) {
			super(option, null);
		}
		protected CLocalOption(String option, String argument) {
			super(option, argument);
		}
		public void send(Session session) throws CVSException {
			session.sendArgument(option);
			if (argument != null) session.sendArgument(argument);
		}
        public CLocalOption[] addTo(CLocalOption[] options) {
            if (this.isElementOf(options)) {
                return options;
            }
            CLocalOption[] newOptions = new CLocalOption[options.length + 1];
            System.arraycopy(options, 0, newOptions, 0, options.length);
            newOptions[options.length] = this;
            return newOptions;
        }
        public CLocalOption[] removeFrom(CLocalOption[] options) {
            if (!this.isElementOf(options)) {
                return options;
            }
            List result = new ArrayList();
            for (int i = 0; i < options.length; i++) {
                CTag.CLocalOption option = options[i];
                if (!option.equals(this)) {
                    result.add(option);
                }
            }
            return (CLocalOption[]) result.toArray(new CLocalOption[result.size()]);
        }
	}
	
	public IStatus execute(Session session, GlobalOption[] globalOptions,
			CLocalOption[] localOptions, CVSTag tag, String[] arguments, ICommandOutputListener listener,
			IProgressMonitor monitor) throws CVSException {
			
			if(tag.getType() != CVSTag.VERSION && tag.getType() != CVSTag.BRANCH) {
				throw new CVSException(new CVSStatus(IStatus.ERROR, CVSMessages.Tag_notVersionOrBranchError)); 
			}
			
			// Add the CREATE_BRANCH option for a branch tag
			if (tag.getType() == CVSTag.BRANCH) {
				if ( ! CREATE_BRANCH.isElementOf(localOptions)) {
					CLocalOption[] newLocalOptions = new CLocalOption[localOptions.length + 1];
					System.arraycopy(localOptions, 0, newLocalOptions, 0, localOptions.length);
					newLocalOptions[newLocalOptions.length - 1] = C_CREATE_BRANCH;
					localOptions = newLocalOptions;
				}
			}
			
			// Add the tag name to the start of the arguments
			String[] newArguments = new String[arguments.length + 1];
			newArguments[0] = tag.getName();
			System.arraycopy(arguments, 0, newArguments, 1, arguments.length);
			
			return executeAction(session, globalOptions, localOptions, newArguments, listener, monitor);	
		}
	
	
	public IStatus execute(Session session, GlobalOption[] globalOptions, CLocalOption[] localOptions, 
			CVSTag tag, ICVSResource[] arguments, ICommandOutputListener listener, IProgressMonitor monitor) 
			throws CVSException {
			
			String[] stringArguments = convertArgumentsForOpenSession(arguments, session);

			return execute(session, globalOptions, localOptions, tag, stringArguments, listener, monitor);
		}

	
	/**
	 * Executes a CVS command.
	 * <p>
	 * Dispatches the commands, retrieves the results, and determines whether or
	 * not an error occurred.  A listener may be supplied to capture message text
	 * that would normally be written to the standard error and standard output
	 * streams of a command line CVS client.
	 * </p>
	 * @param session the open CVS session
	 * @param globalOptions the array of global options, or NO_GLOBAL_OPTIONS
	 * @param localOptions the array of local options, or NO_LOCAL_OPTIONS
	 * @param arguments the array of arguments (usually filenames relative to localRoot), or NO_ARGUMENTS
	 * @param listener the command output listener, or null to discard all messages
	 * @param monitor the progress monitor
	 * @return a status code indicating success or failure of the operation
	 * @throws CVSException if a fatal error occurs (e.g. connection timeout)
	 */
	public IStatus executeAction(final Session session, final GlobalOption[] globalOptions,
		final CLocalOption[] localOptions, final String[] arguments, final ICommandOutputListener listener,
		IProgressMonitor pm) throws CVSException {		
		final IStatus[] status = new IStatus[1];
		ICVSRunnable job = new ICVSRunnable() {
			public void run(IProgressMonitor monitor) throws CVSException {
				// update the global and local options
				GlobalOption[] gOptions = filterGlobalOptions(session, globalOptions);
				CLocalOption[] lOptions = filterLocalOptions(session, gOptions, localOptions);
				
				// print the invocation string to the console
				if (session.isOutputToConsole() || Policy.isDebugProtocol()) {
					IPath commandRootPath;
					IResource resource = session.getLocalRoot().getIResource();
					if (resource == null) {
						commandRootPath = Path.EMPTY;
					} else {
						commandRootPath = resource.getFullPath();
					}
					String line = constructCommandInvocationString(commandRootPath, gOptions, lOptions, arguments);
					ConsoleListeners.getInstance().commandInvoked(session, line);
					if (Policy.isDebugProtocol()) Policy.printProtocolLine("CMD> " + line); //$NON-NLS-1$
				}
				
				// run the command
				try {
				    session.setCurrentCommand(CTag.this);
					status[0] = doExecute(session, gOptions, lOptions, arguments, listener, monitor);
					notifyConsoleOnCompletion(session, status[0], null);
				} catch (CVSException e) {
					notifyConsoleOnCompletion(session, null, e);
					throw e;
				} catch (RuntimeException e) {
					notifyConsoleOnCompletion(session, null, e);
					throw e;
				}
			}
		};
		if (isWorkspaceModification()) {
			session.getLocalRoot().run(job, pm);
		} else {
			job.run(pm);
		}
		return status[0];
	}
	
	private void notifyConsoleOnCompletion(Session session, IStatus status, Exception exception) {
		ConsoleListeners.getInstance().commandCompleted(session, status, exception);
		if (Policy.isDebugProtocol()) {
			if (status != null) Policy.printProtocolLine("RESULT> " + status.toString()); //$NON-NLS-1$
			else Policy.printProtocolLine("RESULT> " + exception.toString()); //$NON-NLS-1$
		}
	}
	
	/**
	 * Allows commands to filter the set of local options to be sent.
	 * Subclasses that override this method should call the superclass.
	 * 
	 * @param session the session
	 * @param globalOptions the global options, read-only
	 * @param localOptions the local options, read-only
	 * @return the filtered local options
	 */
	protected CLocalOption[] filterLocalOptions(Session session, GlobalOption[] globalOptions, CLocalOption[] localOptions) {
		return localOptions;
	}
	
	/**
	 * Constucts the CVS command invocation string corresponding to the arguments.
	 * 
	 * @param globalOptions the global options
	 * @param localOption the local options
	 * @param arguments the arguments
	 * @return the command invocation string
	 */
	private String constructCommandInvocationString(IPath commandRootPath, GlobalOption[] globalOptions,
		CLocalOption[] localOptions, String[] arguments) {
		StringBuffer commandLine = new StringBuffer("cvs"); //$NON-NLS-1$
		for (int i = 0; i < globalOptions.length; ++i) {
			String option = globalOptions[i].toString();
			if (option.length() == 0) continue;
			commandLine.append(' ');
			commandLine.append(option);
		}
		commandLine.append(' ');
		commandLine.append(getRequestId());
		for (int i = 0; i < localOptions.length; ++i) {
			String option = localOptions[i].toString();
			if (option.length() == 0) continue;
			commandLine.append(' ');
			commandLine.append(option);
		}
		for (int i = 0; i < arguments.length; ++i) {
			if (arguments[i].length() == 0) continue;
			commandLine.append(" \""); //$NON-NLS-1$
			IPath completePath = commandRootPath;
			if (!arguments[i].equals(Session.CURRENT_LOCAL_FOLDER)) {
				completePath = completePath.append(arguments[i]);
			}
			commandLine.append(completePath.toString());
			commandLine.append('"'); 
		}
		return commandLine.toString();
	}	
	
	protected IStatus doExecute(Session session, GlobalOption[] globalOptions,
			CLocalOption[] localOptions, String[] arguments, ICommandOutputListener listener,
			IProgressMonitor monitor) throws CVSException {
			ICVSResource[] resources = null;
			/*** setup progress monitor ***/
			monitor = Policy.monitorFor(monitor);
			monitor.beginTask(null, 100);
			Policy.checkCanceled(monitor);
			try {
				/*** prepare for command ***/
				// clear stale command state from previous runs
//				session.setNoLocalChanges(DO_NOT_CHANGE.isElementOf(globalOptions));
//				session.setModTime(null);

				/*** initiate command ***/
				// send global options
				for (int i = 0; i < globalOptions.length; i++) {
					globalOptions[i].send(session);
				}
				Policy.checkCanceled(monitor);
				// send local options
				for (int i = 0; i < localOptions.length; i++) {
					localOptions[i].send(session);
				}
				Policy.checkCanceled(monitor);
				// compute the work resources
				resources = computeWorkResources(session, localOptions, arguments);			
				Policy.checkCanceled(monitor);
				// send local working directory state contributes 48% of work
				resources = sendLocalResourceState(session, globalOptions, localOptions,
						resources, Policy.infiniteSubMonitorFor(monitor, 48));
				Policy.checkCanceled(monitor);
				// escape file names, see bug 149683
				for(int i = 0; i < arguments.length; i++){
					if(arguments[i].startsWith("-")){ //$NON-NLS-1$
						arguments[i] = "./" + arguments[i]; //$NON-NLS-1$
					}
				}
				// send arguments
				sendArguments(session, arguments);
				// send local working directory path
				sendLocalWorkingDirectory(session);

				// if no listener was provided, use the command's default in order to get error reporting
				if (listener == null) listener = getDefaultCommandOutputListener();

				/*** execute command and process responses ***/
				// Processing responses contributes 50% of work.
				IStatus status = executeRequest(session, listener, Policy.subMonitorFor(monitor, 50));

				// Finished adds last 2% of work.
				status = commandFinished(session, globalOptions, localOptions, resources, Policy.subMonitorFor(monitor, 2),
					status);
				return status;
			} finally {			
				monitor.done();
			}
		}
	
	protected ICVSResource[] computeWorkResources(Session session, CLocalOption[] localOptions,
			String[] arguments) throws CVSException {
				
			if (arguments.length < 1) throw new IllegalArgumentException();
			String[] allButFirst = new String[arguments.length - 1];
			System.arraycopy(arguments, 1, allButFirst, 0, arguments.length - 1);

			ICVSFolder localRoot = session.getLocalRoot();

			if (arguments.length == 0) {
				// As a convenience, passing no arguments to the CVS command
				// implies the command will operate on the local root folder.
				return new ICVSResource[] { localRoot };
			} else {
				// Assume all arguments represent resources that are descendants
				// of the local root folder.
				ICVSResource[] resources = new ICVSResource[arguments.length];
				for (int i = 0; i < arguments.length; i++) {
					ICVSResource resource = localRoot.getChild(arguments[i]);				
					// file does not exist, it could have been deleted. It doesn't matter
					// which type we return since only the name of the resource is used
					// and sent to the server.
					if(resource==null) {
						if(localRoot.getName().length()==0) {
							// Return a folder because it is the safest choice when
							// localRoot is a handle to the IWorkspaceRoot!
							resource = localRoot.getFolder(arguments[i]);
						} else {
							resource = localRoot.getFile(arguments[i]);
						}
					}
					resources[i] = resource;
				}
				return resources;
			}
		}
	
	/**
	 * Cleans up after command execution.
	 * [template method] 
	 * <p>
	 * The default implementation is a no-op.  Subclasses may override this
	 * method to follow up command execution on the server with clean up
	 * operations on local resources.
	 * </p>
	 *
	 * @param session the CVS session
	 * @param globalOptions the global options for the command
	 * @param localOptions the local options for the command
	 * @param resources the resource arguments for the command
	 * @param monitor the progress monitor
	 * @param status the status accumulated so far. If the code == CVSStatus.SERVER_ERROR
	 *    then the command failed
	 * @return status the status past in plus any additional status accumulated during the finish
	 */
	protected IStatus commandFinished(Session session, GlobalOption[] globalOptions,
		CLocalOption[] localOptions, ICVSResource[] resources, IProgressMonitor monitor,
		IStatus status) throws CVSException {
			return status;
	}
	
	
	protected ICVSResource[] sendLocalResourceState(Session session, GlobalOption[] globalOptions,
			CLocalOption[] localOptions, ICVSResource[] resources, IProgressMonitor monitor)
			throws CVSException {			

			// Send all folders that are already managed to the server
			if (true) {
				new CusTagFileSender(session, localOptions).visit(session, resources, monitor);
			} else {
		//		new FileStructureVisitor(session, localOptions, false, false).visit(session, resources, monitor);
			}
			return resources;
		}
	
}
